# ISC License
#
# Copyright (c) 2021, Timothée Mazzucotelli
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Adapted from https://github.com/mkdocstrings/mkdocstrings/blob/main/scripts/gen_ref_nav.py
"""Generate the code reference pages and navigation."""

from __future__ import annotations

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

for path in sorted(Path("src").rglob("*.py")):
    module_path = path.relative_to("src").with_suffix("")
    doc_path = path.relative_to("src").with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1].startswith("_"):
        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"::: {ident}")

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Tests for the getter of regression distributions."""

from __future__ import annotations

import pytest

from gemseo_mlearning.active_learning.distributions import KrigingDistribution
from gemseo_mlearning.active_learning.distributions import RegressorDistribution
from gemseo_mlearning.active_learning.distributions import get_regressor_distribution


@pytest.mark.parametrize(
    ("algorithm", "class_"),
    [("linear_algo", RegressorDistribution), ("kriging_algo", KrigingDistribution)],
)
def test_get_regressor_distribution(request, algorithm, class_):
    """Check the getter of the distribution of a regression algorithm."""
    assert isinstance(
        get_regressor_distribution(request.getfixturevalue(algorithm)), class_
    )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Expected improvement."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from numpy import array
from numpy import atleast_2d
from numpy import dot
from numpy import max as np_max
from numpy import maximum
from numpy import mean
from numpy import nan_to_num
from numpy import square
from scipy.stats import norm

from gemseo_mlearning.active_learning.acquisition_criteria.level_set.base_ei_ef import (  # noqa: E501
    BaseEIEF,
)

if TYPE_CHECKING:
    from gemseo.typing import NumberArray

    from gemseo_mlearning.active_learning.distributions.regressor_distribution import (  # noqa: E501
        RegressorDistribution,
    )

warnings.filterwarnings("ignore")


class EI(BaseEIEF):
    r"""The expected improvement.

    This acquisition criterion is expressed as

    $$EI[x]
    =\mathbb{E}\left[\max\left((\kappa\mathbb{S}[Y(x)])^2-(y - Y(x))^2,0\right)\right]$$

    where $y$ is the model output value characterizing the level set
    and $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$.

    In the case of a Gaussian process regressor,
    it has an analytic expression:

    $$
    EI[x] = \mathbb{V}[Y(x)]\times
    (
    (\kappa^2-1-t^2)(\Phi(t^+)-\Phi(t^-))
    -2t(\phi(t^+)-\phi(t^-))
    +t^+\phi(t^+)
    -t^-\phi(t^-)
    )
    $$

    where $\Phi$ and $\phi$ are respectively
    the cumulative and probability density functions
    of the standard normal distribution,
    $t=\frac{y - \mathbb{E}[Y(x)]}{\mathbb{S}[Y(x)]}$,
    $t^+=t+\kappa$,
    $t^-=t-\kappa$
    and $\kappa>0$.

    For the acquisition of $q>1$ points at a time,
    the acquisition criterion changes to

    $$EI[x_1,\dots,x_q] = \mathbb{E}\left[\max_{1\leq i \leq q}\left(
    \max((\kappa\mathbb{S}[Y(x_i)])^2 - (y - Y(x_i))^2,0)
    \right)\right]$$

    where the expectation is taken with respect to the distribution of
    the random vector $(Y(x_1),\dots,Y(x_q))$.
    There is no analytic expression
    and the acquisition is thus instead evaluated with crude Monte-Carlo.
    """

    def _compute(self, input_value: NumberArray) -> NumberArray | float:  # noqa: D102
        # See Proposition 4, Bect et al, 2012
        standard_deviation, t, t_m, t_p = self._get_material(input_value)
        return nan_to_num(
            standard_deviation**2
            * (
                (self._kappa**2 - 1 - t**2) * (norm.cdf(t_p) - norm.cdf(t_m))
                - 2 * t * ((pdf_t_p := norm.pdf(t_p)) - (pdf_t_m := norm.pdf(t_m)))
                + t_p * pdf_t_p
                - t_m * pdf_t_m
            ),
        )

    def _compute_empirically(  # noqa: D102
        self, input_value: NumberArray
    ) -> NumberArray | float:
        # See Proposition 14, Ben Salem et al, 1998
        self._regressor_distribution: RegressorDistribution
        ndim_is_two = input_value.ndim == 2
        input_data = atleast_2d(input_value)
        samples = self._compute_samples(
            input_data=input_data, n_samples=self._compute_samples
        )
        weights = self._regressor_distribution.evaluate_weights(input_data)
        improvement = (
            self._kappa * self._compute_standard_deviation(input_value)
        ) ** 2 - square(self._output_value - samples)
        expected_improvement = maximum(improvement, 0.0)
        expected_improvement = array([
            dot(weights[:, index], expected_improvement[:, index, :])
            for index in range(expected_improvement.shape[1])
        ])
        if ndim_is_two:
            return nan_to_num(expected_improvement)

        return nan_to_num(expected_improvement[0])

    def _compute_by_batch(self, q_input_values: NumberArray) -> NumberArray | float:  # noqa: D102
        q_input_values = self._reshape_input_values(q_input_values)
        try:
            samples = self._compute_samples(
                input_data=q_input_values, n_samples=self._mc_size
            )[..., 0]
            return nan_to_num(
                mean(
                    np_max(
                        maximum(
                            (
                                self._kappa
                                * self._compute_standard_deviation(q_input_values)
                            )
                            ** 2
                            - square(self._output_value - samples.T),
                            0,
                        ),
                        axis=1,
                    ),
                ),
            )
        # distribution is not positive definite.
        except TypeError:
            return 0.0



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# Default settings."""

from __future__ import annotations

from gemseo import configure
from gemseo import configure_logger
from gemseo import sample_disciplines
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from gemseo.uncertainty.statistics.empirical_statistics import EmpiricalStatistics

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.problems.rosenbrock.rosenbrock_discipline import (
    RosenbrockDiscipline,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace

# Update the configuration of |g| to speed up the script (use configure() with care)
configure(False, False, True, False, False, False, False)

configure_logger()

# %%
# The use of active learning methods
# dedicated to quantile estimation
# is illustrated in this example,
# with all default settings.
# The function with the quantile of interest is
# the Rosenbrock function $f(x_1,x_2)=(1-x_1)^2+100(x_2-x_1^2)^2$:
discipline = RosenbrockDiscipline()
# %%
# with $x_1$ and $x_2$ uniformly distributed over $[-2,2]^2$:
uncertain_space = RosenbrockSpace()

# %%
# First,
# we create an initial training dataset using an optimal LHS including 10 samples:
learning_dataset = sample_disciplines(
    [discipline], uncertain_space, "y", algo_name="OT_OPT_LHS", n_samples=10
)

# %%
# and an initial Gaussian process regressor from OpenTURNS:
regressor = OTGaussianProcessRegressor(learning_dataset, trend="quadratic")

# %%
# Then,
# we look for 20 points that will help us
# to approximate the 35% quantile.
# By default,
# for this purpose,
# the active learning algorithm looks
# for the point minimizing the U-function
# with the help of the SLSQP algorithm
# applied in a multistart framework.
level = 0.35
active_learning = ActiveLearningAlgo(
    "Quantile",
    uncertain_space,
    regressor,
    level=level,
    uncertain_space=uncertain_space,
)
active_learning.acquire_new_points(discipline, 20)

# %%
# Then,
# we plot the history of the quantity of interest
active_learning.plot_qoi_history()
# %%
# as well as
# the training points,
# the original model,
# the Gaussian process regressor
# and the U-function
# after the last acquisition:
active_learning.plot_acquisition_view(discipline=discipline)

# %%
# Finally,
# we compare the estimated quantile
# from the active learning procedure
# to the Monte Carlo estimate
# for both algorithms:
dataset = sample_disciplines(
    [discipline], uncertain_space, "y", algo_name="OT_MONTE_CARLO", n_samples=10000
)
reference_quantile = EmpiricalStatistics(dataset, ["y"]).compute_quantile(level)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Base class for acquisition criteria to estimate a level set."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion import (  # noqa: E501
    BaseAcquisitionCriterion,
)

if TYPE_CHECKING:
    from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (  # noqa: E501
        BaseRegressorDistribution,
    )


class BaseLevelSet(BaseAcquisitionCriterion):
    """The base class for acquisition criteria to estimate a level set estimation."""

    _output_value: float
    """The model output value characterizing the level set."""

    def __init__(
        self,
        regressor_distribution: BaseRegressorDistribution,
        output_value: float,
        batch_size: int = 1,
        mc_size: int = 10_000,
    ) -> None:
        """
        Args:
            output_value: The model output value characterizing the level set.
        """  # noqa: D205 D212 D415
        super().__init__(regressor_distribution, batch_size=batch_size, mc_size=mc_size)
        self._output_value = output_value

    def update(self, output_value: float | None = None) -> None:
        """
        Args:
            output_value: The model output value characterizing the level set.
                If ``None``, do not update the acquisition criterion.
        """  # noqa: D205, D212
        super().update()
        if output_value is not None:
            self._output_value = output_value



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Expected feasibility."""

from __future__ import annotations

from gemseo_mlearning.active_learning.acquisition_criteria.level_set.ef import EF as _EF
from gemseo_mlearning.active_learning.acquisition_criteria.quantile.base_quantile import (  # noqa: E501
    BaseQuantile,
)


class EF(BaseQuantile):
    r"""The expected feasibility.

    This acquisition criterion is expressed as

    $$
    EF[x] = \mathbb{E}\left[\max(\kappa\mathbb{S}[Y(x)] - |y_{\alpha} - Y(x)|,0)\right]
    $$

    where $y_{\alpha}$ is the $\alpha$-quantile of the model output
    and $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$.

    In the case of a Gaussian process regressor,
    it has an analytic expression:

    $$
    EF[x] = \mathbb{S}[Y(x)]\times
    \left(
    \kappa
    (\Phi(t^+)-\Phi(t^-))
    -t(2\Phi(t)-\Phi(t^+)-\Phi(t^-))
    -(2\phi(t)-\phi(t^+)-\phi(t^-))
    \right)
    $$

    where $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$,
    $t=\frac{y_{\alpha} - \mathbb{E}[Y(x)]}{\mathbb{S}[Y(x)]}$,
    $t^+=t+\kappa$,
    $t^-=t-\kappa$
    and $\kappa>0$.

    For the acquisition of $q>1$ points at a time,
    the acquisition criterion changes to

    $$EF[x_1,\dots,x_q] = \mathbb{E}\left[\max_{1\leq i\leq q}\left(
    \max(\kappa\mathbb{S}[Y(x_i)] - |y_{\alpha} - Y(x_i)|,0)
    \right)\right]$$

    where the expectation is taken with respect to the distribution of
    the random vector $(Y(x_1),\dots,Y(x_q))$.
    There is no analytic expression
    and the acquisition is thus instead evaluated with crude Monte-Carlo.
    """

    _LEVEL_SET_CLASS = _EF



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# Acquisition algorithm."""

from __future__ import annotations

import matplotlib.pyplot as plt
from gemseo import configure
from gemseo import configure_logger
from gemseo import sample_disciplines
from gemseo.disciplines.surrogate import SurrogateDiscipline
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from numpy import unique

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.problems.rosenbrock.rosenbrock_discipline import (
    RosenbrockDiscipline,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace

# Update the configuration of |g| to speed up the script (use configure() with care)
configure(False, False, True, False, False, False, False)

configure_logger()

# %%
# The use of active learning methods
# dedicated to optimization
# is illustrated in this example.
# More specifically,
# we aim to test here
# the impact of the choice
# of the optimization algorithm
# used to find the next
# acquisition point
# on the active learning procedure.
# The function with the level set of interest is
# the Rosenbrock function $f(x_1,x_2)=(1-x_1)^2+100(x_2-x_1^2)^2$:
discipline = RosenbrockDiscipline()
# %%
# with $x_1$ and $x_2$ belonging to $[-2,2]^2$:
input_space = RosenbrockSpace()

# %%
# First,
# we create an initial training dataset using an optimal LHS including 10 samples:
learning_dataset = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_OPT_LHS", n_samples=10
)

# %%
# and two identical initial
# Gaussian process regressors from OpenTURNS:
regressor_1 = OTGaussianProcessRegressor(
    learning_dataset,
    trend="quadratic",
)
regressor_2 = OTGaussianProcessRegressor(
    learning_dataset,
    trend="quadratic",
)

# %%
# Then,
# we build two active learning algorithms
# to test the impact of the choice
# of acquisition algorithm
# used to optimize the acquisition criterion
# on the active learning procedure.
# One uses the SLSQP gradient-based routine
# in a multistart fashion (default)
# for the optimization
# of the acquisition criterion,
# and the second the NELDER-MEAD gradient-free algorithm,
# also in a multistart fashion.
# All other settings are put to
# their default values.
value_level = 400
active_learning_1 = ActiveLearningAlgo(
    "LevelSet",
    input_space,
    regressor_1,
    output_value=value_level,
)
active_learning_2 = ActiveLearningAlgo(
    "LevelSet",
    input_space,
    regressor_2,
    output_value=value_level,
)
active_learning_1.acquire_new_points(discipline, n_samples=20)
active_learning_2.set_acquisition_algorithm(algo_name="NELDER-MEAD")
active_learning_2.acquire_new_points(discipline, n_samples=20)

# %%
# To study the results,
# for both active learning algorithms,
# we plot the training points,
# the estimated level sets
# alongside the original model.

# Creation of the grid
# and estimation of the different quantities
n_test = 10
surrogate_1 = SurrogateDiscipline(active_learning_1.regressor)
surrogate_2 = SurrogateDiscipline(active_learning_2.regressor)
observations = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_FULLFACT", n_samples=n_test**2
).values
observations_gp_1 = sample_disciplines(
    [surrogate_1], input_space, "y", algo_name="OT_FULLFACT", n_samples=n_test**2
).values
observations_gp_2 = sample_disciplines(
    [surrogate_2], input_space, "y", algo_name="OT_FULLFACT", n_samples=n_test**2
).values

# Plotting the contours of the Rosenbrock function
# alongside the learning points
# and the level sets.
plt.figure()
points_1 = active_learning_1.regressor.learning_set.to_numpy()
points_2 = active_learning_2.regressor.learning_set.to_numpy()
level_set_exact = plt.contour(
    unique(observations[:, 0]),
    unique(observations[:, 1]),
    observations[:, 2].reshape(n_test, n_test),
    levels=[value_level],
    colors="red",
)
level_set_gp_1 = plt.contour(
    unique(observations_gp_1[:, 0]),
    unique(observations_gp_1[:, 1]),
    observations_gp_1[:, 2].reshape(n_test, n_test),
    levels=[value_level],
    colors="tab:blue",
    linestyles="dotted",
)
level_set_gp_2 = plt.contour(
    unique(observations_gp_2[:, 0]),
    unique(observations_gp_2[:, 1]),
    observations_gp_2[:, 2].reshape(n_test, n_test),
    levels=[value_level],
    colors="tab:orange",
    linestyles="dotted",
)
plt.clabel(level_set_exact, levels=[value_level], fontsize=10, colors="red")
plt.annotate("Target level set", (-0.2, 0.75), color="red")
plt.annotate(
    "Level set estimated with the multistart SLSQP", (-1.5, 0.5), color="tab:blue"
)
plt.annotate(
    "Level set estimated with the multistart NELDER-MEAD",
    (-1.5, 0.25),
    color="tab:orange",
)
plt.contour(
    unique(observations[:, 0]),
    unique(observations[:, 1]),
    observations[:, 2].reshape(n_test, n_test),
)
bar = plt.colorbar()
bar.set_label(r"$f(x_1,x_2)$")
plt.scatter(
    points_1[:, 0],
    points_1[:, 1],
    marker="*",
    label="Learning points from algo with multistart SLSQP",
)
plt.scatter(
    points_2[:, 0],
    points_2[:, 1],
    marker="*",
    label="Learning points from algo with NELDER-MEAD",
)
plt.xlabel(r"$x_1$")
plt.ylabel(r"$x_2$")
plt.legend()
plt.show()
# The estimated level set
# estimated with both acquisition algorithm
# provide a good approximation of the target.
# Furthermore,
# though the learning points are not strictly
# speaking the same,
# several are located close the other
# if not the same.



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The Branin function as a discipline."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gemseo.core.discipline.discipline import Discipline
from numpy import array
from numpy import concatenate

from gemseo_mlearning.problems.branin.functions import compute_gradient
from gemseo_mlearning.problems.branin.functions import compute_output

if TYPE_CHECKING:
    from collections.abc import Iterable

    from gemseo.typing import StrKeyMapping


class BraninDiscipline(Discipline):
    """The Branin function as a discipline."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__()
        self.input_grammar.update_from_names(["x1", "x2"])
        self.output_grammar.update_from_names(["y"])
        self.default_input_data.update({
            name: array([0.0]) for name in self.input_grammar.names
        })

    def _run(self, input_data: StrKeyMapping) -> None:
        inputs_array = concatenate([
            self.io.data[name] for name in self.io.input_grammar
        ])
        self.io.update_output_data({"y": array([compute_output(inputs_array)])})

    def _compute_jacobian(
        self,
        input_names: Iterable[str] = (),
        output_names: Iterable[str] = (),
    ) -> None:
        inputs_array = concatenate([
            self.io.data[name] for name in self.io.input_grammar
        ])
        self.jac = {
            "y": {
                input_name: array([[derivative]])
                for input_name, derivative in zip(
                    self.io.input_grammar.names,
                    compute_gradient(inputs_array),
                )
            }
        }



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# Surrogate-based optimization using in-house features."""

from __future__ import annotations

from gemseo import configure_logger
from gemseo import execute_algo
from gemseo.post.dataset.zvsxy import ZvsXY
from gemseo.problems.dataset.rosenbrock import create_rosenbrock_dataset
from gemseo.problems.optimization.rosenbrock import Rosenbrock

configure_logger()

# %%
# In this example,
# we seek to minimize the Rosenbrock function $f(x,y)=(1-x)^2+100(y-x^2)^2$
# over the design space $[-2,2]^2$.
# First,
# we instantiate the problem with $(0, 0)$ as initial guess:
problem = Rosenbrock()

# %%
# Then,
# we minimize the Rosenbrock function using:
#
# - the `"SBO"` algorithm,
# - a maximum number of 40 evaluations,
#   including the initial one at the center of the design space
#   (this first point is common to all optimization algorithms)
#   and the initial training dataset,
# - its default settings,
#   namely:
#
#   - the expected improvement as acquisition criterion,
#   - 1 point acquired at a time,
#   - the `GaussianProcessRegressor` based on scikit-learn,
#   - 10 initial training points
#     based on an optimized latin hypercube sampling (LHS) technique,
#   - a multi-start local optimization of the acquisition criterion
#     from 50 start points with a limit of 20 iterations per local optimization.
execute_algo(problem, algo_name="SBO", max_iter=40)

# %%
# We can see
# that the solution is close to the theoretical one $(x^*,f^*)=((1,1),0)$.
#
# We can also visualize all the evaluations
# and note that most of the points have been added in the valley as expected:
optimization_history = problem.to_dataset()

initial_point = optimization_history[0:1]
initial_point.name = "Initial point"

initial_training_points = optimization_history[1:12]
initial_training_points.name = "Initial training points"

acquired_points = optimization_history[12:]
acquired_points.name = "Acquired points"

visualization = ZvsXY(
    create_rosenbrock_dataset(900),
    ("x", 0),
    ("x", 1),
    "rosen",
    fill=False,
    other_datasets=(initial_point, initial_training_points, acquired_points),
)
visualization.execute(save=False, show=True)

# %%
# Lastly,
# we can compare the solution to the one obtained with COBYLA,
# which is another popular gradient-free optimization algorithm:
execute_algo(Rosenbrock(), algo_name="NLOPT_COBYLA", max_iter=40)

# %%
# and conclude that for this problem and this initial guess,
# the surrogate-based algorithm is better than COBYLA.



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import pytest
from numpy import array
from numpy import ndarray
from numpy.testing import assert_equal

from gemseo_mlearning.problems.branin.branin_discipline import BraninDiscipline
from gemseo_mlearning.problems.branin.functions import compute_gradient
from gemseo_mlearning.problems.branin.functions import compute_output


@pytest.fixture(scope="module")
def input_values() -> tuple[dict[str, ndarray], ndarray]:
    """The input values of interest."""
    return {name: array([1.0]) for name in ["x1", "x2"]}, array([1.0, 1.0])


@pytest.fixture(scope="module")
def discipline() -> BraninDiscipline:
    """The Branin discipline."""
    return BraninDiscipline()


def test_init(discipline) -> None:
    """Check the instantiation of the discipline."""
    assert discipline.name == "BraninDiscipline"
    input_names = ["x1", "x2"]
    assert list(discipline.input_grammar.names) == input_names
    assert list(discipline.output_grammar.names) == ["y"]
    assert discipline.default_input_data == {name: array([0.0]) for name in input_names}


def test_execute(discipline, input_values) -> None:
    """Check the output value of the discipline."""
    assert discipline.execute(input_values[0])["y"][0] == compute_output(
        input_values[1]
    )


def test_gradient(discipline, input_values) -> None:
    """Check the gradient of the discipline."""
    gradient = discipline.linearize(
        input_data=input_values[0], compute_all_jacobians=True
    )["y"]
    assert_equal(
        array([gradient[name][0, 0] for name in discipline.input_grammar.names]),
        compute_gradient(input_values[1]),
    )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A class for surrogate-based optimization."""

from __future__ import annotations

import logging
import pickle
from pathlib import Path
from typing import TYPE_CHECKING

from gemseo.algos.doe.factory import DOELibraryFactory
from gemseo.algos.hashable_ndarray import HashableNdarray
from gemseo.datasets.io_dataset import IODataset
from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor
from gemseo.mlearning.regression.algos.factory import RegressorFactory
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from gemseo.utils.constants import READ_ONLY_EMPTY_DICT
from gemseo.utils.logging_tools import LoggingContext
from numpy import hstack
from numpy import newaxis
from pandas import concat

from gemseo_mlearning.active_learning.acquisition_criteria.minimum.minimum import (
    Minimum,
)
from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo

if TYPE_CHECKING:
    from collections.abc import Mapping

    from gemseo.algos.base_driver_library import DriverLibrarySettingType
    from gemseo.algos.optimization_problem import OptimizationProblem
    from gemseo.mlearning.core.algos.ml_algo import MLAlgoParameterType


class SurrogateBasedOptimizer:
    """An optimizer based on surrogate models."""

    __STOP_BECAUSE_ALREADY_KNOWN = "The acquired input data is already known."
    __STOP_BECAUSE_MAX_ACQUISITIONS = "All the data acquisitions have been made."

    __active_learning_algo: ActiveLearningAlgo
    """The active learning algorithm to acquire new samples to learn."""

    __dataset: IODataset
    """The original learning dataset enriched by new samples."""

    __initial_input_samples: tuple[HashableNdarray, ...]
    """The initial input samples, if any."""

    __regression_file_path: str | Path
    """The path to the file to save the regression model.

    If empty, do not save the regression model.
    """

    __problem: OptimizationProblem
    """The optimization problem."""

    def __init__(
        self,
        problem: OptimizationProblem,
        acquisition_algorithm: str,
        doe_size: int = 0,
        doe_algorithm: str = "OT_OPT_LHS",
        doe_settings: Mapping[str, DriverLibrarySettingType] = READ_ONLY_EMPTY_DICT,
        regression_algorithm: str | BaseRegressor = OTGaussianProcessRegressor.__name__,
        regression_settings: Mapping[str, MLAlgoParameterType] = READ_ONLY_EMPTY_DICT,
        regression_file_path: str | Path = "",
        **acquisition_settings: DriverLibrarySettingType,
    ) -> None:
        """
        Args:
            acquisition_algorithm: The name of the algorithm to optimize the data
                acquisition criterion.
                N.B. this algorithm must handle integers if some of the optimization
                variables are integers.
            problem: The optimization problem.
            doe_size: Either the size of the initial DOE
                or 0 if the size is inferred from doe_settings.
                This argument is ignored
                when regression_algorithm is a
                [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            doe_algorithm: The name of the algorithm for the initial sampling.
                This argument is ignored
                when regression_algorithm is a
                [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            doe_settings: The settings of the algorithm for the initial sampling.
                This argument is ignored
                when regression_algorithm is a
                [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            regression_algorithm: Either the name of the regression algorithm
                approximating the objective function over the design space
                or the regression algorithm itself.
            regression_settings: The settings of the regression algorithm.
                If transformer is missing,
                use :attr:`.BaseMLRegressionAlgo.DEFAULT_TRANSFORMER`.
                This argument is ignored
                when regression_algorithm is a
                [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            regression_file_path: The path to the file to save the regression model.
                If empty, do not save the regression model.
            **acquisition_settings: The settings of the algorithm to optimize
                the data acquisition criterion.
        """  # noqa: D205, D212, D415
        self.__problem = problem
        database = problem.database
        self.__initial_input_samples = tuple(database.keys())
        if isinstance(regression_algorithm, BaseRegressor):
            self.__dataset = regression_algorithm.learning_set
        else:
            # Store max_iter as it will be overwritten by DOELibrary
            max_iter = problem.evaluation_counter.maximum
            settings = dict(doe_settings)
            if doe_size > 0 and "n_samples" not in settings:
                settings["n_samples"] = doe_size

            # Store the listeners as they will be cleared by DOELibrary.
            new_iter_listeners, store_listeners = database.clear_listeners()
            with LoggingContext(logging.getLogger("gemseo")):
                DOELibraryFactory().execute(
                    problem, algo_name=doe_algorithm, **settings
                )

            for listener in new_iter_listeners:
                database.add_new_iter_listener(listener)

            for listener in store_listeners:
                database.add_store_listener(listener)

            self.__dataset = problem.to_dataset(opt_naming=False)
            if self.__initial_input_samples:
                self.__dataset = self.__dataset[len(self.__initial_input_samples) :]

            regression_settings_ = {"transformer": {"inputs": "MinMaxScaler"}}
            regression_settings_.update(dict(regression_settings))
            regression_algorithm = RegressorFactory().create(
                regression_algorithm,
                self.__dataset,
                **regression_settings_,
            )
            # Add the first iteration to the current_iter reset by DOELibrary.
            problem.evaluation_counter.current += 1
            # And restore max_iter.
            problem.evaluation_counter.maximum = max_iter

        self.__active_learning_algo = ActiveLearningAlgo(
            Minimum.__name__, problem.design_space, regression_algorithm
        )
        if acquisition_algorithm:
            self.__active_learning_algo.set_acquisition_algorithm(
                acquisition_algorithm, **acquisition_settings
            )
        self.__regression_file_path = regression_file_path

    def execute(self, number_of_acquisitions: int) -> str:
        """Execute the surrogate-based optimization.

        Args:
            number_of_acquisitions: The number of learning points to be acquired.

        Returns:
            The termination message.
        """
        regressor_distribution = self.__active_learning_algo.regressor_distribution
        regressor_distribution.learn()
        message = self.__STOP_BECAUSE_MAX_ACQUISITIONS
        for _ in range(number_of_acquisitions):
            input_data = self.__active_learning_algo.find_next_point()[0]
            hashed_input_data = HashableNdarray(input_data)
            if hashed_input_data in self.__problem.database and (
                hashed_input_data not in self.__initial_input_samples
            ):
                message = self.__STOP_BECAUSE_ALREADY_KNOWN
                break

            output_data = self.__problem.evaluate_functions(
                design_vector=input_data, design_vector_is_normalized=False
            )[0]
            extra_learning_set = IODataset()
            variable_names_to_n_components = regressor_distribution.algo.sizes
            extra_learning_set.add_group(
                group_name=IODataset.INPUT_GROUP,
                data=input_data[newaxis],
                variable_names=regressor_distribution.input_names,
                variable_names_to_n_components=variable_names_to_n_components,
            )
            output_names = regressor_distribution.output_names
            extra_learning_set.add_group(
                group_name=IODataset.OUTPUT_GROUP,
                data=hstack([output_data[output_name] for output_name in output_names])[
                    newaxis
                ],
                variable_names=output_names,
                variable_names_to_n_components=variable_names_to_n_components,
            )
            self.__dataset = concat(
                [regressor_distribution.algo.learning_set, extra_learning_set],
                ignore_index=True,
            )
            self.__dataset = self.__dataset.map(lambda x: x.real)
            regressor_distribution.change_learning_set(self.__dataset)
            self.__active_learning_algo.update_problem()

            if self.__regression_file_path:
                with Path(self.__regression_file_path).open("wb") as file:
                    pickle.dump(regressor_distribution.algo, file)

        return message



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Lower confidence bound (LCB)."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar

from gemseo_mlearning.active_learning.acquisition_criteria.maximum._confidence_bound import (  # noqa: E501
    ConfidenceBound,
)
from gemseo_mlearning.active_learning.acquisition_criteria.minimum.base_minimum import (
    BaseMinimum,
)

if TYPE_CHECKING:
    from gemseo_mlearning.active_learning.distributions import BaseRegressorDistribution


class LCB(ConfidenceBound, BaseMinimum):
    r"""The lower confidence bound (LCB).

    This acquisition criterion is expressed as

    $$M[x;\kappa] = \mathbb{E}[Y(x)] - \kappa \times \mathbb{S}[Y(x)]$$

    where $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$
    and $\kappa>0$.
    """

    MAXIMIZE: ClassVar[bool] = False

    def __init__(  # noqa: D107
        self,
        regressor_distribution: BaseRegressorDistribution,
        kappa: float = 2.0,
        batch_size: int = 1,
        mc_size: int = 10_000,
    ) -> None:  # noqa: D102
        super().__init__(
            regressor_distribution,
            kappa=-abs(kappa),
            batch_size=batch_size,
            mc_size=mc_size,
        )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Base class for families of acquisition criteria."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar

from gemseo.core.base_factory import BaseFactory

if TYPE_CHECKING:
    from gemseo_mlearning.active_learning.acquisition_criteria.base_factory import (
        BaseAcquisitionCriterionFactory,
    )


class BaseAcquisitionCriterionFamily:
    """The base class for families of acquisition criteria."""

    ACQUISITION_CRITERION_FACTORY: ClassVar[type[BaseAcquisitionCriterionFactory]]


class AcquisitionCriterionFamilyFactory(BaseFactory):
    """The factory of families of acquisition criteria."""

    _CLASS = BaseAcquisitionCriterionFamily
    _PACKAGE_NAMES = ("gemseo_mlearning.active_learning.acquisition_criteria",)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import re

import pytest
from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor
from gemseo.post.dataset.lines import Lines
from gemseo.utils.testing.helpers import image_comparison

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.active_learning.visualization.qoi_history_view import (
    QOIHistoryView,
)


@pytest.fixture(scope="module")
def qoi_history_view(active_learning_algo) -> QOIHistoryView:
    """The QOI history view."""
    return QOIHistoryView(active_learning_algo)


def test_return_type(qoi_history_view):
    """Check the return type of QOIHistoryView."""
    assert isinstance(qoi_history_view.draw(show=False), Lines)


@image_comparison(["default"])
def test_default(qoi_history_view):
    """Check QOIHistoryView with default settings."""
    qoi_history_view.draw(show=False)


@image_comparison(["custom"])
def test_custom(qoi_history_view):
    """Check QOIHistoryView with custom settings."""
    qoi_history_view.draw(show=False, plot_abscissa_variable=True, label="foo")


def test_error_no_qoi(learning_dataset, input_space):
    """Check that an error is raised when there is no QOI."""
    regressor = GaussianProcessRegressor(learning_dataset)
    active_learning_algo = ActiveLearningAlgo("Exploration", input_space, regressor)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "There is no quantity of interest "
            "associated with the acquisition criterion Variance."
        ),
    ):
        QOIHistoryView(active_learning_algo)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import re

import pytest
from numpy import array
from numpy.testing import assert_almost_equal

from gemseo_mlearning.active_learning.acquisition_criteria.minimum.ei import EI
from gemseo_mlearning.active_learning.acquisition_criteria.minimum.lcb import LCB
from gemseo_mlearning.active_learning.acquisition_criteria.minimum.mean import Mean


@pytest.mark.parametrize(
    ("cls", "kwargs", "input_value", "expected"),
    [
        (EI, {}, array([0.0]), array([0.0])),
        (LCB, {}, array([0.0]), array([1.0])),
        (LCB, {"kappa": 3.0}, array([0.0]), array([1.0])),
        (Mean, {}, array([0.0]), array([1.0])),
        (EI, {}, array([[0.0]] * 2), array([[0.0]] * 2)),
        (LCB, {}, array([[0.0]] * 2), array([[1.0]] * 2)),
        (LCB, {"kappa": 3.0}, array([[0.0]] * 2), array([[1.0]] * 2)),
        (Mean, {}, array([[0.0]] * 2), array([[1.0]] * 2)),
    ],
)
def test_minimum_kriging_regressor(
    kriging_distribution, cls, kwargs, input_value, expected
):
    """Check the criteria deriving from BaseMinimum with a Kriging distribution."""
    criterion = cls(kriging_distribution, **kwargs)
    assert_almost_equal(criterion.func(input_value), expected)


@pytest.mark.parametrize(
    ("cls", "kwargs", "input_value", "expected"),
    [
        (EI, {}, array([0.123]), array([0.75])),
        (LCB, {}, array([0.123]), array([-0.620])),
        (LCB, {"kappa": 3.0}, array([0.123]), array([-1.12])),
        (Mean, {}, array([0.123]), array([0.37])),
        (EI, {}, array([[0.123]] * 2), array([[0.75]] * 2)),
        (LCB, {}, array([[0.123]] * 2), array([[-0.62]] * 2)),
        (LCB, {"kappa": 3.0}, array([[0.123]] * 2), array([[-1.12]] * 2)),
        (Mean, {}, array([[0.123]] * 2), array([[0.37]] * 2)),
    ],
)
def test_minimum_regressor(algo_distribution, cls, kwargs, input_value, expected):
    """Check the criteria deriving from BaseMinimum with a non-Kriging distribution."""
    criterion = cls(algo_distribution, **kwargs)
    assert_almost_equal(criterion.func(input_value), expected, decimal=2)


@pytest.mark.parametrize(
    ("cls", "input_value"),
    [
        (EI, array([0.123])),
        (LCB, array([0.123])),
        (Mean, array([0.123])),
    ],
)
def test_bad_parallel_regressor(algo_distribution, cls, input_value):
    """Check that parallelized criteria with non GP regressor lead to failure."""
    criterion = cls(algo_distribution, batch_size=2)

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "Parallelization with batch_size > 1 is not yet implemented "
            "for regressors that are not based on a random process."
        ),
    ):
        criterion.func(input_value)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Settings for the surrogate-based optimization algorithm."""

from __future__ import annotations

from collections.abc import Mapping  # noqa: TC003
from pathlib import Path  # noqa: TC003

from gemseo.algos.base_driver_library import DriverSettingType  # noqa: TC002
from gemseo.algos.opt.base_optimizer_settings import (  # noqa: TC002
    BaseOptimizerSettings,
)
from gemseo.mlearning.core.algos.ml_algo import MLAlgoSettingsType  # noqa: TC002
from gemseo.mlearning.regression.algos.base_regressor import (  # noqa: TC002
    BaseRegressor,
)
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from pydantic import Field
from pydantic import NonNegativeInt
from pydantic import PositiveInt
from strenum import StrEnum


class AcquisitionCriterion(StrEnum):
    r"""An acquisition criterion.

    In the following,
    the training output values already used
    and the random output of the surrogate model at a given input point $x$
    are respectively denoted $\{y_1,\ldots,y_n\}$ and $Y(x)$.
    The expectation and the standard deviation of $Y(x)$ are respectively denoted
    $\mathbb{E}[Y(x)]$ and $\mathbb{S}[Y(x)]$.
    """

    EI = "EI"
    r"""The expected improvement.

    The acquisition criterion is $\mathbb{E}[\max(\min(y_1,\dots,y_n)-Y(x),0]$.
    """

    CB = "CB"
    r"""The confidence bound.

    The acquisition criterion is $\mathbb{E}[Y(x)]-3\mathbb{S}[Y(x)]$.
    """

    Output = "Output"
    r"""The mean output.

    The acquisition criterion is $\mathbb{E}[Y(x)]$.
    """


class SBO_Settings(BaseOptimizerSettings):  # noqa: N801
    """The settings for the surrogate-based optimization algorithm."""

    acquisition_algorithm: str = Field(
        default="",
        description=(
            """The name of the algorithm to optimize the data acquisition criterion.
            If empty, use the default algorithm with its default settings."""
        ),
    )

    acquisition_settings: Mapping[str, DriverSettingType] = Field(
        default_factory=dict,
        description=(
            """The settings of the algorithm
            to optimize the data acquisition criterion.
            Ignored when `acquisition_algorithm` is empty."""
        ),
    )

    batch_size: PositiveInt = Field(
        default=1, description="The number of points to be acquired in parallel."
    )

    criterion: AcquisitionCriterion = Field(
        default=AcquisitionCriterion.EI, description="The acquisition criterion."
    )

    doe_algorithm: str = Field(
        default="OT_OPT_LHS",
        description=(
            """The name of the DOE algorithm for the initial sampling.
            This argument is ignored
            when regression_algorithm is a
            [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            """
        ),
    )

    doe_settings: Mapping[str, DriverSettingType] = Field(
        default_factory=dict,
        description=(
            """The settings of the DOE algorithm for the initial sampling.
            This argument is ignored
            when regression_algorithm is a
            [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            """
        ),
    )

    doe_size: NonNegativeInt = Field(
        default=10,
        description=(
            """Either the initial DOE size or 0 if it is inferred from `doe_settings`.
            This argument is ignored
            when regression_algorithm is a
            [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            """
        ),
    )

    mc_size: PositiveInt = Field(
        default=10_000,
        description="The sample size to estimate the acquisition criteria in parallel.",
    )

    normalize_design_space: bool = Field(
        default=False,
        description=(
            """Whether to normalize the design space variables between 0 and 1."""
        ),
    )

    regression_algorithm: str | BaseRegressor = Field(
        default=OTGaussianProcessRegressor.__name__,
        description=(
            """The regression algorithm.
            Either the name of the regression algorithm
            approximating the objective function over the design space
            or the regression algorithm itself.
            """
        ),
    )

    regression_file_path: str | Path = Field(
        default="",
        description=(
            """The path to the file to save the regression model.
            If empty, do not save the regression model.
            """
        ),
    )

    regression_settings: Mapping[str, MLAlgoSettingsType] = Field(
        default_factory=dict,
        description=(
            """The settings of the regression algorithm.
            This argument is ignored
            when regression_algorithm is a
            [BaseRegressor][gemseo.mlearning.regression.algos.base_regressor.BaseRegressor].
            """
        ),
    )


# TODO: API: remove this alias.
SBOSettings = SBO_Settings



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Regressor distributions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gemseo.mlearning.regression.algos.base_random_process_regressor import (
    BaseRandomProcessRegressor,
)

from gemseo_mlearning.active_learning.distributions.kriging_distribution import (
    KrigingDistribution,
)
from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
    RegressorDistribution,
)

if TYPE_CHECKING:
    from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor

    from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (  # noqa: E501
        BaseRegressorDistribution,
    )


def get_regressor_distribution(
    regressor: BaseRegressor,
    use_bootstrap: bool = True,
    use_loo: bool = False,
    size: int | None = None,
) -> BaseRegressorDistribution:
    """Return the distribution of a regressor.

    Args:
        regressor: The regression algorithm.
        use_bootstrap: Whether to use bootstrap for resampling.
            If `False`, use cross-validation.
        use_loo: Whether to use leave-one-out resampling when
            `use_bootstrap` is `False`.
            If `False`, use parameterized cross-validation.
        size: The size of the resampling set,
            i.e. the number of times the regression algorithm is rebuilt.
            If `None`,
            [N_BOOTSTRAP][gemseo_mlearning.active_learning.distributions.regressor_distribution.RegressorDistribution.N_BOOTSTRAP]
            in the case of bootstrap
            and
            [N_FOLDS][gemseo_mlearning.active_learning.distributions.regressor_distribution.RegressorDistribution.N_FOLDS]
            in the case of cross-validation.
            This argument is ignored in the case of leave-one-out.

    Returns:
        The distribution of the regression algorithm.
    """
    if isinstance(regressor, BaseRandomProcessRegressor):
        return KrigingDistribution(regressor)

    return RegressorDistribution(regressor, use_bootstrap, use_loo, size)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Base class for acquisition criteria."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar

from gemseo.core.mdo_functions.mdo_function import MDOFunction
from numpy import ones

from gemseo_mlearning.active_learning.distributions.kriging_distribution import (  # noqa: E501
    KrigingDistribution,
)

if TYPE_CHECKING:
    from gemseo.mlearning.core.algos.ml_algo import DataType
    from gemseo.typing import NumberArray

    from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (  # noqa: E501
        BaseRegressorDistribution,
    )


class BaseAcquisitionCriterion(MDOFunction):
    """Base class for acquisition criteria."""

    output_range: float
    """The output range."""

    _compute_mean: Callable[[DataType], DataType]
    """The function to compute the output mean at a given input point."""

    _compute_standard_deviation: Callable[[DataType], DataType]
    """The function to compute the output standard deviation at a given input point."""

    _compute_variance: Callable[[DataType], DataType]
    """The function to compute the output variance at a given input point."""

    _qoi: Any
    """The quantity of interest."""

    _regressor_distribution: BaseRegressorDistribution
    """The distribution of the regressor."""

    _batch_size: int
    """The number of points to be acquired in parallel; if `1`, acquire points
    sequentially."""

    _mc_size: int
    """The sample size to estimate the acquisition criteria in parallel."""

    MAXIMIZE: ClassVar[bool] = True
    """Whether this acquisition criterion must be maximized."""

    def __init__(
        self,
        regressor_distribution: BaseRegressorDistribution,
        batch_size: int = 1,
        mc_size: int = 10_000,
    ) -> None:
        """
        Args:
            regressor_distribution: The distribution of the regressor.
            batch_size: The number of points to be acquired in parallel;
                if `1`, acquire points sequentially.
            mc_size: The sample size to estimate the acquisition criterion in parallel.
        """  # noqa: D205 D212 D415
        self._compute_mean = regressor_distribution.compute_mean
        self._compute_standard_deviation = (
            regressor_distribution.compute_standard_deviation
        )
        self._compute_variance = regressor_distribution.compute_variance
        output_data = regressor_distribution.learning_set.output_dataset.to_numpy()
        self.output_range = output_data.max() - output_data.min()
        self._regressor_distribution = regressor_distribution
        try:
            jac = self._compute_jacobian(
                ones(regressor_distribution.algo.input_dimension)
            )
        except NotImplementedError:
            jac = None

        self._batch_size = batch_size
        self._mc_size = mc_size
        self._compute_samples = regressor_distribution.compute_samples

        if batch_size == 1:
            if isinstance(regressor_distribution, KrigingDistribution):
                compute_output = self._compute
            else:
                compute_output = self._compute_empirically
        else:
            if isinstance(regressor_distribution, KrigingDistribution):
                compute_output = self._compute_by_batch
            else:
                compute_output = self._compute_by_batch_empirically

        self._qoi = None
        super().__init__(
            compute_output,
            self.__class__.__name__,
            jac=jac,
        )

        self.update()

    @property
    def qoi(self) -> Any:
        """The quantity of interest."""
        return self._qoi

    @abstractmethod
    def _compute(self, input_value: NumberArray) -> NumberArray | float:
        """Compute the acquisition criterion value using a random process regressor.

        Args:
            input_value: The model input value.

        Returns:
            The acquisition criterion value.
        """

    def _compute_empirically(self, input_value: NumberArray) -> NumberArray | float:
        """Compute the acquisition criterion value using resampling.

        Args:
            input_value: The model input value.

        Returns:
            The acquisition criterion value.
        """
        return self._compute(input_value)

    def _compute_by_batch(self, q_input_values: NumberArray) -> NumberArray | float:
        """Compute the parallelized acquisition criterion value using a random process.

        Args:
            q_input_values: The batch of $q$ input values.

        Returns:
            The parallelized acquisition criterion value.
        """
        msg = "Parallelization is not defined for this acquisition criterion."
        raise NotImplementedError(msg)

    def _compute_by_batch_empirically(
        self, q_input_values: NumberArray
    ) -> NumberArray | float:
        """Compute the parallelized acquisition criterion using resampling.

        Args:
            q_input_values: The batch of $q$ input values.

        Returns:
            The parallelized acquisition criterion value.
        """
        msg = (
            "Parallelization with batch_size > 1 is not yet implemented "
            "for regressors that are not based on a random process."
        )
        raise NotImplementedError(msg)

    def _compute_jacobian(self, input_value: NumberArray) -> NumberArray:
        """Compute the Jacobian of the acquisition criterion.

        Args:
            input_value: The model input value.

        Returns:
            The Jacobian of the acquisition criterion.

        Raises:
            NotImplementedError: When the function is not implemented.
        """
        raise NotImplementedError

    @property
    def _scaling_factor(self) -> float:
        """The factor to scale values in the output space."""
        if self.output_range == 0:
            return 1.0

        return self.output_range

    def _reshape_input_values(self, input_values: NumberArray) -> NumberArray:
        r"""Reshape an array from `(q\times d,)` to `(q, d)`.

        Args:
            input_values: The input values shaped as `(q\times d,)`.

        Returns:
            The input values shaped as `(q, d)`.
        """
        return input_values.reshape(self._batch_size, -1)

    def update(self) -> None:
        """Update the acquisition criterion."""
        data = self._regressor_distribution.learning_set.output_dataset.to_numpy()
        self.output_range = data.max() - data.min()

    def __truediv__(self, other: MDOFunction | float) -> MDOFunction:
        new_criterion = super().__truediv__(other)
        new_criterion._regressor_distribution = self._regressor_distribution
        new_criterion.output_range = self.output_range
        return new_criterion

    def __neg__(self) -> MDOFunction:
        new_criterion = super().__neg__()
        new_criterion._regressor_distribution = self._regressor_distribution
        new_criterion.output_range = self.output_range
        return new_criterion

    def __add__(self, other: MDOFunction | float) -> MDOFunction:
        new_criterion = super().__add__(other)
        new_criterion._regressor_distribution = self._regressor_distribution
        new_criterion.output_range = self.output_range
        return new_criterion

    def __sub__(self, other: MDOFunction | float) -> MDOFunction:
        new_criterion = super().__sub__(other)
        new_criterion._regressor_distribution = self._regressor_distribution
        new_criterion.output_range = self.output_range
        return new_criterion

    def __mul__(self, other: MDOFunction | float) -> MDOFunction:
        new_criterion = super().__mul__(other)
        new_criterion._regressor_distribution = self._regressor_distribution
        new_criterion.output_range = self.output_range
        return new_criterion



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    AUTHORS:
#       - Francois Gallard
"""The efficient global optimization (EGO) algorithm of SMT."""

from __future__ import annotations

from math import isfinite
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Final

from gemseo.algos.design_space_utils import get_value_and_bounds
from gemseo.algos.opt.base_optimization_library import BaseOptimizationLibrary
from gemseo.algos.opt.base_optimization_library import OptimizationAlgorithmDescription
from gemseo.algos.optimization_result import OptimizationResult
from numpy import atleast_2d
from smt.applications.ego import EGO
from smt.surrogate_models import GEKPLS
from smt.surrogate_models import GPX
from smt.surrogate_models import KPLS
from smt.surrogate_models import KPLSK
from smt.surrogate_models import KRG
from smt.surrogate_models import MGP
from smt.utils.design_space import DesignSpace as SMTDesignSpace

from gemseo_mlearning.algos.opt.smt._parallel_evaluator import ParallelEvaluator
from gemseo_mlearning.algos.opt.smt.ego_settings import SMT_EGO_Settings

if TYPE_CHECKING:
    from gemseo.algos.optimization_problem import OptimizationProblem
    from gemseo.typing import RealArray
    from smt.surrogate_models.surrogate_model import SurrogateModel


class SMTEGO(BaseOptimizationLibrary):
    """Surrogate-based optimizers from SMT."""

    __NAMES_TO_CLASSES: Final[dict[str, type[SurrogateModel]]] = {
        "GEKPLS": GEKPLS,
        "GPX": GPX,
        "KPLS": KPLS,
        "KPLSK": KPLSK,
        "KRG": KRG,
        "MGP": MGP,
    }
    """The names of the surrogate models bound to the SMT classes."""

    ALGORITHM_INFOS: ClassVar[dict[str, OptimizationAlgorithmDescription]] = {
        "SMT_EGO": OptimizationAlgorithmDescription(
            library_name="SMT",
            algorithm_name="SMT_EGO",
            description="Efficient Global Optimization",
            internal_algorithm_name="SMT_EGO",
            website="https://smt.readthedocs.io/en/latest/_src_docs/applications/ego.html",  # noqa: E501
            Settings=SMT_EGO_Settings,
        )
    }

    def __init__(self, algo_name: str = "SMT_EGO") -> None:  # noqa: D107
        super().__init__(algo_name)

    def _run(
        self, problem: OptimizationProblem, **settings: Any
    ) -> tuple[None, None, RealArray, RealArray, RealArray]:
        design_space = problem.design_space
        x_0, lower_bounds, upper_bounds = get_value_and_bounds(
            design_space, normalize_ds=self._normalize_ds
        )
        surrogate = settings["surrogate"]
        if isinstance(surrogate, str):
            smt_design_space = SMTDesignSpace(
                atleast_2d(
                    list(
                        zip(
                            [
                                lower_bound if isfinite(lower_bound) else None
                                for lower_bound in lower_bounds
                            ],
                            [
                                upper_bound if isfinite(upper_bound) else None
                                for upper_bound in upper_bounds
                            ],
                        )
                    )
                )
            )
            surrogate = self.__NAMES_TO_CLASSES[surrogate](
                design_space=smt_design_space, print_global=False
            )

        n_parallel = settings["n_parallel"]
        evaluator = ParallelEvaluator(n_parallel)

        max_iter = settings["max_iter"]
        n_doe = settings["n_doe"]
        n_iter = max_iter - n_doe - 1
        if n_iter <= 0:
            msg = (
                f"max_iter must be strictly greater than n_doe+1 = {n_doe + 1}; "
                f"got {max_iter}."
            )
            raise ValueError(msg)

        ego = EGO(
            surrogate=surrogate,
            n_doe=n_doe,
            n_iter=n_iter,
            criterion=settings["criterion"],
            n_parallel=n_parallel,
            qEI=settings["qEI"],
            n_start=settings["n_start"],
            evaluator=evaluator,
            random_state=settings["random_state"],
            n_max_optim=settings["n_max_optim"],
        )
        x_opt, y_opt, _, _, _ = ego.optimize(fun=problem.objective.evaluate)
        if self._normalize_ds:
            x_opt = design_space.unnormalize_vect(x_opt)
            x_0 = design_space.unnormalize_vect(x_0)
        return (
            None,
            None,
            x_0,
            x_opt,
            y_opt,
        )

    def _get_result(
        self,
        problem: OptimizationProblem,
        message: Any,
        status: Any,
        x_0: RealArray,
        x_opt: RealArray,
        f_opt: RealArray,
    ) -> OptimizationResult:
        return OptimizationResult(
            x_0=x_0,
            x_opt=x_opt,
            f_opt=f_opt,
            optimizer_name=self.algo_name,
            n_obj_call=problem.objective.n_calls,
            is_feasible=True,
        )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
r"""The Branin use case to benchmark and illustrate active learning algorithms.

The Branin function
$$f(x_1,x_2) =
\left(15x_2 - \frac{5.1}{4\pi^2}(15x_1-5)^2 + \frac{5}{\pi}(15x_1-5)-6\right)^2
+ \left(10- \frac{10}{8\pi}\right)\cos(15x_1-5)
+10$$
is commonly studied through the random variable $Y=f(X_1,X_2)$
where $X_1$ and $X_2$ are independent random variables
uniformly distributed over $[0,1]$.

See [@molga2005test].
"""

from __future__ import annotations



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""An acquisition plot."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gemseo.utils.matplotlib_figure import save_show_figure
from matplotlib import pyplot as plt
from numpy import array
from numpy import linspace
from numpy import meshgrid
from numpy import zeros

if TYPE_CHECKING:
    from pathlib import Path

    from gemseo.core.discipline.discipline import Discipline
    from gemseo.typing import RealArray
    from matplotlib.figure import Figure

    from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo


class AcquisitionView:
    """View of points acquired during active learning.

    This visualization tool only works with a 2-dimensional input space.
    """

    __algo: ActiveLearningAlgo
    """The active learning algorithm."""

    __input_dimension: int
    """The input dimension."""

    def __init__(self, active_learning_algo: ActiveLearningAlgo) -> None:
        """
        Args:
            active_learning_algo: The active learning algorithm
                using sequential acquisition.

        Raises:
            NotImplementedError: When the active learning algorithm uses parallel
                acquisition.
        """  # noqa: D205, D212
        if active_learning_algo.batch_size > 1:
            msg = (
                "AcquisitionView does not support active learning algorithm using "
                "parallel acquisition."
            )
            raise NotImplementedError(msg)

        self.__algo = active_learning_algo
        self.__input_dimension = (
            active_learning_algo.regressor_distribution.algo.input_dimension
        )

    def draw(
        self,
        new_point: RealArray | None = None,
        discipline: Discipline | None = None,
        filled: bool = True,
        n_test: int = 30,
        show: bool = True,
        file_path: str | Path = "",
    ) -> Figure:
        """Draw the points acquired through active learning.

        This visualization includes four surface plots
        representing variables of interest in function of the two inputs:

        - (top left) the surface plot of the discipline,
        - (top right) the surface plot of the acquisition criterion,
        - (bottom left) the surface plot of the regressor,
        - (bottom right) the standard deviation of the regressor.

        The $N$ data used to initialize the regressor are represented by small dots,
        the point optimizing the acquisition criterion is represented by a big dot
        (this will be the next point to learn)
        and the points added by the active learning procedure are represented
        by plus signs and labeled by their position in the learning dataset.

        Args:
            new_point: The new point to be acquired.
            discipline: The discipline for which `n_test**2` evaluations will be done.
                If `None`,
                do not plot the discipline, i.e. the first subplot.
                In particular,
                if the discipline is costly,
                it is better to leave this argument to `None`.
            n_test: The number of points per dimension.
            filled: Whether to plot filled contours.
                Otherwise, plot contour lines.
            show: Whether to display the view.
            file_path: The file path to save the view.
                If empty, do not save it.

        Returns:
            The acquisition view.
        """
        # Create grid.
        input_space = self.__algo.input_space
        lower_bounds = input_space.get_lower_bounds()
        upper_bounds = input_space.get_upper_bounds()
        test_x1 = linspace(lower_bounds[0], upper_bounds[0], n_test)
        test_x2 = linspace(lower_bounds[1], upper_bounds[1], n_test)
        grid = array(meshgrid(test_x1, test_x2)).T.reshape(-1, 2)

        # Generate data.
        distribution = self.__algo.regressor_distribution
        final_dataset = distribution.algo.learning_set
        #    The learning input samples.
        points = final_dataset.input_dataset.to_numpy()
        points_x = points[:, 0]
        points_y = points[:, 1]
        #    The predictions, the standard deviations and the criterion values.
        predictions = distribution.predict(grid).reshape((n_test, n_test)).T
        std = distribution.compute_standard_deviation(grid).reshape((n_test, n_test)).T
        acquisition_criterion = self.__algo.acquisition_criterion.original.func
        criterion_values = acquisition_criterion(grid).reshape((n_test, n_test)).T
        x_name, y_name = final_dataset.input_names
        output_name = final_dataset.output_names[0]
        #    The observations if the discipline is available.
        observations = None
        if discipline is not None:
            observations = zeros((n_test, n_test))
            for i in range(n_test):
                for j in range(n_test):
                    xij = array([test_x1[j], test_x2[i]])
                    input_data = {x_name: array([xij[0]]), y_name: array([xij[1]])}
                    observations[i, j] = discipline.execute(input_data)[output_name][0]

        # Create figure and sub-figures.
        fig, axes = plt.subplots(2, 2)
        titles = [
            ["Discipline", self.__algo.acquisition_criterion.__class__.__name__],
            ["Surrogate", "Standard deviation"],
        ]
        data = [[observations, criterion_values], [predictions, std]]
        cf = []
        color = "white" if filled else "black"
        n_initial_samples = self.__algo.n_initial_samples
        contour_method = "contourf" if filled else "contour"
        for i in range(2):
            for j in range(2):
                if (i, j) == (0, 0) and discipline is None:
                    continue

                ax = axes[i, j]
                cf.append(getattr(ax, contour_method)(test_x1, test_x2, data[i][j]))
                for x, y in zip(
                    points_x[:n_initial_samples],
                    points_y[:n_initial_samples],
                ):
                    ax.plot(x, y, ".", ms=2, color=color)

                for index, (x, y) in enumerate(
                    zip(
                        points_x[n_initial_samples:],
                        points_y[n_initial_samples:],
                    )
                ):
                    ax.plot(x, y, "+", color=color)
                    ax.annotate(
                        str(n_initial_samples + 1 + index),
                        (0.05 + x, 0.05 + y),
                        color=color,
                    )

                if new_point is not None:
                    ax.plot(*new_point, "o", color=color)

                ax.set_title(titles[i][j])

        if discipline is None:
            fig.delaxes(axes.flatten()[0])
            fig.colorbar(cf[0], ax=axes[0, 1])
            axes[0, 1].set_xticks([])
            fig.colorbar(cf[1], ax=axes[1, 0])
            fig.colorbar(cf[2], ax=axes[1, 1])
            axes[1, 1].set_yticks([])
        else:
            fig.colorbar(cf[0], ax=axes[:, 0])
            axes[0, 0].set_xticks([])
            fig.colorbar(cf[1], ax=axes[0, 1])
            axes[0, 1].set_xticks([])
            axes[0, 1].set_yticks([])
            fig.colorbar(cf[3], ax=axes[1, 1])
            axes[1, 1].set_yticks([])

        save_show_figure(fig, show, file_path)
        return fig



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The uncertain space used in the Branin use case."""

from __future__ import annotations

from openturns import Uniform

from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace


def test_rosenbrock_space() -> None:
    """Check the Rosenbrock space."""
    space = RosenbrockSpace()
    assert space.dimension == 2
    assert space.variable_names == ["x1", "x2"]
    for distribution in space.distributions.values():
        assert len(distribution.marginals) == 1
        distribution = distribution.marginals[0].distribution
        assert distribution.getParameter() == (-2.0, 2.0)
        assert isinstance(distribution, Uniform)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Base class for regressor distributions."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from gemseo.mlearning.data_formatters.regression_data_formatters import (
    RegressionDataFormatters,
)
from gemseo.utils.metaclasses import ABCGoogleDocstringInheritanceMeta

if TYPE_CHECKING:
    from gemseo.datasets.dataset import Dataset
    from gemseo.datasets.io_dataset import IODataset
    from gemseo.mlearning.core.algos.ml_algo import DataType
    from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor
    from gemseo.typing import NumberArray


class BaseRegressorDistribution(metaclass=ABCGoogleDocstringInheritanceMeta):
    """The distribution of a regressor."""

    algo: BaseRegressor
    """The regression model."""

    _samples: list[int]
    """The indices of the learning samples in the learning dataset."""

    _transform_input_group: bool
    """Whether to transform the input group."""

    _transform_output_group: bool
    """Whether to transform the output group."""

    _input_variables_to_transform: list[str]
    """The names of the input variables to be transformed."""

    _output_variables_to_transform: list[str]
    """The names of the output variables to be transformed."""

    def __init__(self, regressor: BaseRegressor) -> None:
        """
        Args:
            regressor: A regression model.
        """  # noqa: D205 D212 D415
        self.algo = regressor
        self._samples = []
        self._transform_input_group = self.algo._transform_input_group
        self._transform_output_group = self.algo._transform_output_group
        self._input_variables_to_transform = self.algo._input_variables_to_transform
        self._output_variables_to_transform = self.algo._output_variables_to_transform

    @property
    def learning_set(self) -> IODataset:
        """The learning dataset used by the regressor."""
        return self.algo.learning_set

    @property
    def input_names(self) -> list[str]:
        """The input names of the regressor."""
        return self.algo.input_names

    @property
    def output_names(self) -> list[str]:
        """The output names of the regressor."""
        return self.algo.output_names

    @property
    def output_dimension(self) -> int:
        """The output dimension of the regressor."""
        return self.algo.output_dimension

    def learn(self, samples: list[int] | None = None) -> None:
        """Train the regressor from the learning dataset.

        Args:
            samples: The indices of the learning samples.
                If `None`, use the whole learning dataset
        """
        self._samples = samples or range(len(self.learning_set))
        self.algo.learn(self._samples)

    def predict(
        self,
        input_data: DataType,
    ) -> DataType:
        """Compute the output value(s) of the regressor.

        The user can specify the input data either as a NumPy array,
        e.g. `array([1., 2., 3.])`
        or as a dictionary,
        e.g. `{"a": array([1.]), "b": array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.

        Returns:
            The output value(s) of the regressor.
        """
        return self.algo.predict(input_data)

    @abstractmethod
    def compute_confidence_interval(
        self,
        input_data: DataType,
        level: float = 0.95,
    ) -> (
        tuple[
            dict[str, NumberArray],
            dict[str, NumberArray],
            tuple[NumberArray, NumberArray],
        ]
        | None
    ):
        """Compute the lower bounds and upper bounds of the regressor.

        The user can specify the input data either as a NumPy array,
        e.g. `array([1., 2., 3.])`
        or as a dictionary,
        e.g.  `{"a": array([1.]), "b": array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.
            level: A quantile level.

        Returns:
            The lower and upper bound values of the regressor.
        """

    @abstractmethod
    def compute_mean(
        self,
        input_data: DataType,
    ) -> DataType:
        """Compute the output mean of the regressor.

        The user can specify the input data either as a NumPy array,
        e.g. `array([1., 2., 3.])`
        or as a dictionary,
        e.g.  `{"a": array([1.]), "b": array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.

        Returns:
            The output mean of the regressor.
        """

    @abstractmethod
    def compute_variance(
        self,
        input_data: DataType,
    ) -> DataType:
        """Compute the output variance of the regressor.

        The user can specify the input data either as a NumPy array,
        e.g. `array([1., 2., 3.])`
        or as a dictionary,
        e.g.  `{"a": array([1.]), "b": array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.

        Returns:
            The output variance of the regressor.
        """

    @RegressionDataFormatters.format_dict
    @RegressionDataFormatters.format_samples()
    def compute_standard_deviation(
        self,
        input_data: DataType,
    ) -> DataType:
        """Compute the output standard deviation of the regressor.

        The user can specify the input data either as a NumPy array,
        e.g. `array([1., 2., 3.])`
        or as a dictionary,
        e.g.  `{"a": array([1.]), "b": array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.

        Returns:
            The output standard deviation  of the regressor.
        """
        return self.compute_variance(input_data) ** 0.5

    def change_learning_set(self, learning_set: Dataset) -> None:
        """Re-train the regressor from a new learning set.

        Args:
            learning_set: The new learning set.
        """
        self.algo.learning_set = learning_set
        self.learn()

    @abstractmethod
    def compute_samples(
        self,
        input_data: NumberArray,
        n_samples: int,
    ) -> NumberArray:
        """Generate samples from the random process.

        Args:
            input_data: The $N$ input points of dimension $d$
                at which to observe the random process;
                shaped as `(N, d)`.
            n_samples: The number of samples `M`.

        Returns:
            The output samples per output dimension shaped as `(N, M, p)`
            where `p` is the output dimension.
        """



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from numpy import array

from gemseo_mlearning.active_learning.distributions.kriging_distribution import (
    KrigingDistribution,
)

if TYPE_CHECKING:
    from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor


@pytest.fixture(scope="module")
def distribution(kriging_algo: GaussianProcessRegressor) -> KrigingDistribution:
    """The distribution of the Kriging regression model."""
    distribution = KrigingDistribution(kriging_algo)
    distribution.learn()
    return distribution


def test_init(kriging_algo):
    """Check the initialization of the distribution."""
    assert KrigingDistribution(kriging_algo).algo == kriging_algo


@pytest.mark.parametrize(
    "point",
    [
        array([0.0]),
        array([0.5]),
        array([1.0]),
        array([[0.0], [0.5]]),
        {"x": array([0.0])},
        {"x": array([[0.0], [0.5]])},
    ],
)
def test_compute_mean(distribution, point):
    """Check the computation of the mean."""
    result = distribution.compute_mean(point)
    if isinstance(point, dict):
        result = result["y"]
        point = point["x"]

    assert result.shape == point.shape
    assert result[0] == pytest.approx(1 if point[0] in [0.0, 1.0] else 0.0, abs=1e-10)


@pytest.mark.parametrize(
    "point",
    [
        array([0.0]),
        array([0.5]),
        array([1.0]),
        array([[0.0], [0.5]]),
        {"x": array([0.0])},
        {"x": array([[0.0], [0.5]])},
    ],
)
def test_compute_variance(distribution, point):
    """Check the computation of the variance."""
    result = distribution.compute_variance(point)
    if isinstance(point, dict):
        point = point["x"]
        result = result["y"]
    assert result.shape == point.shape
    assert result[0] == pytest.approx(0, abs=1e-3)


@pytest.mark.parametrize(
    "point",
    [
        array([0.0]),
        array([0.5]),
        array([1.0]),
        array([[0.0], [0.5]]),
        {"x": array([0.0])},
        {"x": array([[0.0], [0.5]])},
    ],
)
def test_confidence_interval(distribution, point):
    """Check the computation of a confidence interval."""
    lower, upper = distribution.compute_confidence_interval(point)
    if isinstance(point, dict):
        point = point["x"]
        lower = lower["y"]
        upper = upper["y"]
    assert (lower < upper).all()
    assert lower.shape == upper.shape == point.shape


@pytest.mark.parametrize(
    "input_data",
    [
        array([1 / 3.97]),
        array([[1 / 3.97], [1 / 1.97]]),
    ],
)
def test_compute_samples(distribution, input_data):
    """Check that the samples from the GP are of appropriate shape."""
    samples = distribution.compute_samples(input_data=input_data, n_samples=1000)
    expected_shape = (
        (1000, 1) if len(input_data.shape) == 1 else (1000, input_data.shape[0], 1)
    )
    assert samples.shape == expected_shape



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Universal regressor distribution."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Callable
from typing import ClassVar
from typing import Final

from gemseo.mlearning.data_formatters.regression_data_formatters import (
    RegressionDataFormatters,
)
from gemseo.utils.data_conversion import concatenate_dict_of_arrays_to_array
from numpy import array
from numpy import array_split
from numpy import atleast_2d
from numpy import delete as npdelete
from numpy import dot
from numpy import exp
from numpy import ones
from numpy import quantile
from numpy import stack
from numpy import sum as npsum
from numpy import unique
from numpy.random import default_rng
from scipy.spatial.distance import euclidean

from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (
    BaseRegressorDistribution,
)

if TYPE_CHECKING:
    from gemseo.datasets.dataset import Dataset
    from gemseo.mlearning.core.algos.ml_algo import DataType
    from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor
    from gemseo.typing import NumberArray


class RegressorDistribution(BaseRegressorDistribution):
    """Universal regressor distribution."""

    method: str
    """The resampling method."""

    size: int
    """The size of the resampling set."""

    weights: list[Callable[[NumberArray], float]]
    """The weight functions related to the sub-algorithms.

    A weight function computes a weight from an input data array.
    """

    N_BOOTSTRAP: ClassVar[int] = 100
    """The default number of replicates for the bootstrap method."""

    N_FOLDS: ClassVar[int] = 5
    """The default number of folds for the cross-validation method."""

    CROSS_VALIDATION: Final[str] = "cv"
    BOOTSTRAP: Final[str] = "b"
    LOO: Final[str] = "loo"

    def __init__(
        self,
        algo: BaseRegressor,
        bootstrap: bool = True,
        loo: bool = False,
        size: int | None = None,
    ) -> None:
        """
        Args:
            bootstrap: The resampling method.
                If `True`, use bootstrap resampling.
                Otherwise, use cross-validation resampling.
            loo: The leave-One-Out sub-method, when bootstrap is `False`.
                If `False`, use parameterized cross-validation,
                Otherwise use leave-one-out.
            size: The size of the resampling set,
                i.e. the number of times the regression algorithm is rebuilt.
                If `None`,
                [N_BOOTSTRAP][gemseo_mlearning.active_learning.distributions.regressor_distribution.RegressorDistribution.N_BOOTSTRAP]
                in the case of bootstrap
                and
                [N_FOLDS][gemseo_mlearning.active_learning.distributions.regressor_distribution.RegressorDistribution.N_FOLDS]
                in the case of cross-validation.
                This argument is ignored in the case of leave-one-out.
        """  # noqa: D205 D212 D415
        if bootstrap:
            self.method = self.BOOTSTRAP
            self.size = size or self.N_BOOTSTRAP
        else:
            if loo:
                self.method = self.LOO
                self.size = len(algo.learning_set)
            else:
                self.method = self.CROSS_VALIDATION
                self.size = size or self.N_FOLDS
        self.algos = [
            algo.__class__(algo.learning_set, settings_model=algo._settings)
            for _ in range(self.size)
        ]
        self.weights = []
        super().__init__(algo)

    def learn(  # noqa: D102
        self,
        samples: list[int] | None = None,
    ) -> None:
        self.weights = []
        super().learn(samples)
        if self.method in [self.CROSS_VALIDATION, self.LOO]:
            n_folds = self.size
            folds = array_split(self._samples, n_folds)
        for index, algo in enumerate(self.algos):
            if self.method == self.BOOTSTRAP:
                new_samples = unique(
                    default_rng(1).choice(self._samples, len(self._samples))
                )
                other_samples = list(set(self._samples) - set(new_samples))
                self.weights.append(self.__weight_function(other_samples))
            else:
                fold = folds[index]
                new_samples = npdelete(self._samples, fold)
                other_samples = list(set(self._samples) - set(new_samples))
                self.weights.append(self.__weight_function(other_samples))

            algo.learn(new_samples.tolist())

    def __weight_function(self, indices: list[int]) -> Callable[[NumberArray], float]:
        """Return a function evaluating the weights at an input vector.

        The weights are w.r.t. the input vectors of the samples
        skipped at iteration indices.

        Args:
            indices: The indices of the samples
                removed from the learning dataset during the training phase.

        Returns:
            The weight function returning a weight from a 1D input array.
        """
        dat = self.learning_set.get_view(
            group_names=self.learning_set.INPUT_GROUP
        ).to_numpy()
        all_indices = set(self._samples)
        rho = max(
            min(euclidean(dat[id1], dat[id2]) for id2 in all_indices - {id1})
            for id1 in all_indices
        )
        in_grp = self.learning_set.INPUT_GROUP

        def weight(
            input_data: NumberArray,
        ) -> float:
            """Weight function returning a weight when a 1D input array is passed.

            Args:
                input_data: A 1D input array.

            Returns:
                A weight related to the input array.
            """
            if isinstance(input_data, Mapping):
                input_data = concatenate_dict_of_arrays_to_array(
                    input_data, self.input_names
                )
            only_one_element = input_data.ndim == 1
            input_data = atleast_2d(input_data)
            distance = ones(input_data.shape[0])
            learning_data = self.learning_set.get_view(group_names=in_grp).to_numpy()
            learning_data = learning_data[indices]
            for learning_datum in learning_data:
                for index, input_datum in enumerate(input_data):
                    distance[index] *= 1 - exp(
                        -(euclidean(learning_datum, input_datum) ** 2) / rho**2
                    )
            if only_one_element:
                distance = distance[0]
            return distance

        return weight

    def predict_members(self, input_data: DataType) -> DataType:
        """Predict the output value with the different machine learning algorithms.

        After prediction, the method stacks the results.

        Args:
            input_data: The input data,
                specified as either as a NumPy array or as dictionary of NumPy arrays
                indexed by inputs names.
                The NumPy array can be either a `(d,)` array
                representing a sample in dimension `d`,
                or a `(M, d)` array representing `M` samples in dimension `d`.

        Returns:
            The output data (dimension `p`) of `N` machine learning algorithms.
                If `input_data.shape == (d,)`, then `output_data.shape == (N, p)`.
                If `input_data.shape == (M,d)`, then `output_data.shape == (N,M,p)`.
        """
        predictions = [algo.predict(input_data) for algo in self.algos]
        if isinstance(input_data, Mapping):
            return {
                name: stack([prediction[name] for prediction in predictions])
                for name in predictions[0]
            }
        return stack(predictions)

    def compute_confidence_interval(  # noqa: D102
        self,
        input_data: DataType,
        level: float = 0.95,
    ) -> (
        tuple[dict[str, NumberArray], dict[str, NumberArray]]
        | tuple[NumberArray, NumberArray]
    ):
        level = (1.0 - level) / 2.0
        predictions = self.predict_members(input_data)
        if isinstance(predictions, Mapping):
            lower = {
                name: quantile(value, level, axis=0)
                for name, value in predictions.items()
            }
            upper = {
                name: quantile(value, 1 - level, axis=0)
                for name, value in predictions.items()
            }
        else:
            lower = quantile(predictions, level, axis=0)
            upper = quantile(predictions, 1 - level, axis=0)
        return lower, upper

    def evaluate_weights(self, input_data: NumberArray) -> NumberArray:
        """Evaluate weights.

        Args:
            input_data: The input data with shape (size, n_input_data)

        Returns:
            The weights.
        """
        weights = array([func(input_data) for func in self.weights])
        return weights / npsum(weights, 0)

    @RegressionDataFormatters.format_dict
    @RegressionDataFormatters.format_samples()
    def compute_mean(self, input_data: DataType) -> DataType:  # noqa: D102
        predictions = self.predict_members(input_data)
        weights = self.evaluate_weights(input_data)
        return self.__average(weights, predictions)

    @RegressionDataFormatters.format_dict
    @RegressionDataFormatters.format_samples()
    def compute_variance(self, input_data: DataType) -> DataType:  # noqa: D102
        predictions = self.predict_members(input_data)
        weights = self.evaluate_weights(input_data)
        term1 = self.__average(weights, predictions**2)
        term2 = self.__average(weights, predictions) ** 2
        return term1 - term2

    def compute_samples(  # noqa: D102
        self,
        input_data: NumberArray,
        n_samples: int,
    ) -> NumberArray:
        return self.predict_members(input_data)

    @staticmethod
    def __average(
        weights: NumberArray,
        data: NumberArray,
    ) -> NumberArray:
        """Return averaged data.

        Args:
            weights: The weights.
            data: The data from which to get the average.

        Returns:
            The averaged value.
        """
        return array([
            dot(weights[:, index], data[:, index, :]) for index in range(data.shape[1])
        ])

    def change_learning_set(self, learning_set: Dataset) -> None:  # noqa: D102
        for algo in self.algos:
            algo.learning_set = learning_set
        super().change_learning_set(learning_set)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Base class for acquisition criteria to estimate a quantile."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from gemseo.algos.parameter_space import ParameterSpace
from gemseo.utils.string_tools import pretty_str
from numpy import atleast_1d
from numpy import quantile

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion import (  # noqa: E501
    BaseAcquisitionCriterion,
)

if TYPE_CHECKING:
    from gemseo.typing import NumberArray
    from gemseo.typing import RealArray

    from gemseo_mlearning.active_learning.acquisition_criteria.level_set.base_level_set import (  # noqa: E501
        BaseLevelSet,
    )
    from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (  # noqa: E501
        BaseRegressorDistribution,
    )


class BaseQuantile(BaseAcquisitionCriterion):
    """The base class for acquisition criteria to estimate a quantile."""

    @property
    @abstractmethod
    def _LEVEL_SET_CLASS(self) -> type[BaseLevelSet]:  # noqa: N802
        """The acquisition criterion to estimate a level set."""

    __level: float
    """The quantile level."""

    __input_data: RealArray
    """The input samples to estimate the quantile."""

    def __init__(
        self,
        regressor_distribution: BaseRegressorDistribution,
        level: float,
        uncertain_space: ParameterSpace,
        n_samples: int = 10_0000,
        batch_size: int = 1,
        mc_size: int = 10_000,
    ) -> None:
        """
        Args:
            level: The quantile level.
            uncertain_space: The uncertain variable space.
            n_samples: The number of samples
                to estimate the quantile of the regressor by Monte Carlo.
        """  # noqa: D205 D212 D415
        input_names = regressor_distribution.input_names
        missing_names = set(input_names) - set(uncertain_space.variable_names)
        if missing_names:
            msg = (
                "The probability distributions of the input variables "
                f"{pretty_str(missing_names, use_and=True)} are missing."
            )
            raise ValueError(msg)

        # Create a new uncertain space sorted by model inputs.
        new_uncertain_space = ParameterSpace()
        new_uncertain_space.add_variables_from(uncertain_space, *input_names)
        self.__input_data = new_uncertain_space.compute_samples(n_samples)
        self.__level = level
        # The value 0. will be replaced by the quantile estimation at each update,
        # including the first one due to super().__init__.
        self.__level_set_criterion = self._LEVEL_SET_CLASS(
            regressor_distribution, 0.0, batch_size=batch_size, mc_size=mc_size
        )

        super().__init__(regressor_distribution, batch_size=batch_size, mc_size=mc_size)

    def update(self) -> None:  # noqa: D102
        super().update()
        qoi = self._qoi = atleast_1d(self.__compute_quantile())
        self.__level_set_criterion.update(output_value=qoi[0])

    def __compute_quantile(self) -> float:
        """Return the quantile estimation.

        Returns:
            The quantile estimation.
        """
        output_data = self._regressor_distribution.predict(self.__input_data)
        return quantile(output_data, self.__level)

    def _compute(self, input_value: NumberArray) -> NumberArray:
        return self.__level_set_criterion.func(input_value)

    def _compute_empirically(self, input_value: NumberArray) -> NumberArray:
        return self.__level_set_criterion.func(input_value)

    def _compute_by_batch(self, input_value: NumberArray) -> NumberArray:
        return self.__level_set_criterion.func(input_value)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from gemseo.algos.opt.base_optimizer_settings import BaseOptimizerSettings
from gemseo.core.base_factory import BaseFactory
from gemseo.mlearning.core.algos.ml_algo_settings import BaseMLAlgoSettings

import gemseo_mlearning.settings.mlearning as mlearning
import gemseo_mlearning.settings.opt as opt

if TYPE_CHECKING:
    from gemseo.algos.base_algorithm_settings import BaseAlgorithmSettings


def get_setting_classes(
    BaseSettings: type[BaseAlgorithmSettings],  # noqa: N803
    package_name: str,
    module_,
) -> list[str]:
    """Return the settings classes given a type of algorithms.

    Args:
        BaseSettings: The base class specific to the type of algorithms.
        package_name: The name of the package.
        module_: The module of settings.

    Returns:
        The settings classes.
    """

    class SettingsFactory(BaseFactory):
        _CLASS = BaseSettings
        _PACKAGE_NAMES = (package_name,)

        @property
        def classes(self) -> list[str]:
            return [
                self.get_class(name)
                for name in super().class_names
                if not name.startswith("Base")
            ]

    for cls in SettingsFactory().classes:
        if cls.__module__.startswith("gemseo_mlearning"):
            yield module_, cls


@pytest.mark.parametrize(
    "module_and_cls",
    get_setting_classes(BaseOptimizerSettings, "gemseo_mlearning.algos.opt", opt),
)
def test_opt_settings(module_and_cls):
    """Check aliases for optimizer settings."""
    module, cls = module_and_cls
    assert cls in module.__dict__.values()


@pytest.mark.parametrize(
    "module_and_cls",
    get_setting_classes(BaseMLAlgoSettings, "gemseo_mlearning.regression", mlearning),
)
def test_machine_learning_settings(module_and_cls):
    """Check aliases for machine learning algorithm settings."""
    module, cls = module_and_cls
    assert cls in module.__dict__.values()



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Upper confidence bound (UCB)."""

from __future__ import annotations

from gemseo_mlearning.active_learning.acquisition_criteria.maximum._confidence_bound import (  # noqa: E501
    ConfidenceBound,
)
from gemseo_mlearning.active_learning.acquisition_criteria.maximum.base_maximum import (
    BaseMaximum,
)


class UCB(ConfidenceBound, BaseMaximum):
    r"""The upper confidence bound (UCB).

    This acquisition criterion is expressed as

    $$M[x;\kappa] = \mathbb{E}[Y(x)] + \kappa \times \mathbb{S}[Y(x)]$$

    where $Y$ is the random process modelling the uncertainty of the surrogate model
    $\hat{f}$ and $\kappa>0$.
    """



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# Non-GP regressor."""

from __future__ import annotations

from gemseo import configure
from gemseo import configure_logger
from gemseo import sample_disciplines
from gemseo.mlearning.regression.algos.rbf import RBFRegressor

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
    RegressorDistribution,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_discipline import (
    RosenbrockDiscipline,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace

# Update the configuration of |g| to speed up the script (use configure() with care)
configure(False, False, True, False, False, False, False)

configure_logger()

# %%
# The use of active learning methods
# dedicated to level set estimation
# is illustrated in this example,
# with all default settings.
# The function with the level set of interest is
# the Rosenbrock function $f(x_1,x_2)=(1-x_1)^2+100(x_2-x_1^2)^2$:
discipline = RosenbrockDiscipline()
# %%
# with $x_1$ and $x_2$ belonging to $[-2,2]^2$:
input_space = RosenbrockSpace()

# %%
# First,
# we create an initial training dataset using an optimal LHS including 10 samples:
learning_dataset = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_OPT_LHS", n_samples=10
)

# %%
# and a universal regressor, namely a radial basis function network based on SciPy:
regressor = RBFRegressor(learning_dataset)
regressor_distribution = RegressorDistribution(regressor, bootstrap=False)
regressor_distribution.learn()

# %%
# Then,
# we look for 20 points that will help us
# to approximate the level-set associated to the 35% quantile.
# By default,
# for this purpose,
# the active learning algorithm looks
# for the point minimizing the U-function
# with the help of the SLSQP gradient-based algorithm
# applied in a multistart framework.
level_value = 300.0
active_learning = ActiveLearningAlgo(
    "LevelSet",
    input_space,
    regressor_distribution,
    output_value=level_value,
)
active_learning.acquire_new_points(discipline, 20)

# %%
# Lastly,
# we plot the training points,
# the original model,
# the RBF regressor
# and the U-function
# after the last acquisition:
active_learning.plot_acquisition_view(discipline=discipline)
# It can be seen
# that the learning points
# are distributed around
# the target level set,
# thus approximating it properly.



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The mean-based criterion to approximate a maximum."""

from __future__ import annotations

from typing import ClassVar

from gemseo_mlearning.active_learning.acquisition_criteria.maximum._mean import (
    Mean as _Mean,
)
from gemseo_mlearning.active_learning.acquisition_criteria.minimum.base_minimum import (
    BaseMinimum,
)


class Mean(_Mean, BaseMinimum):
    r"""Mean-based criterion.

    This acquisition criterion is expressed as

    $$E[x] = \mathbb{E}[Y(x)]$$

    where $Y$ is the random process modelling the uncertainty of the surrogate model
    $\hat{f}$.
    """

    MAXIMIZE: ClassVar[bool] = False



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import pytest
from gemseo.mlearning.regression.algos.linreg import LinearRegressor
from numpy import array
from numpy import exp
from numpy import quantile

from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
    RegressorDistribution,
)


def __weight_func(value, indices):
    terms = [
        1 - exp(-((value - 0.0) ** 2) / 0.5**2),
        1 - exp(-((value - 0.5) ** 2) / 0.5**2),
        1 - exp(-((value - 1.0) ** 2) / 0.5**2),
    ]
    result = 1
    for index in indices:
        result *= terms[index]
    return result


@pytest.fixture(scope="module")
def distribution(linear_algo) -> RegressorDistribution:
    """The distribution of the linear regression model."""
    distribution = RegressorDistribution(linear_algo, bootstrap=False, loo=True)
    distribution.learn()
    return distribution


@pytest.mark.parametrize(
    ("bootstrap", "loo", "size"),
    [
        (True, False, None),
        (True, False, 3),
        (False, False, None),
        (False, False, 3),
        (False, True, None),
    ],
)
def test_init(linear_algo, bootstrap, loo, size):
    """Check the initialization of the distribution."""
    distribution = RegressorDistribution(linear_algo, bootstrap, loo, size)
    if bootstrap:
        assert distribution.method == distribution.BOOTSTRAP
        assert distribution.size == size or distribution.N_BOOTSTRAP
    elif loo:
        assert distribution.method == distribution.LOO
        assert distribution.size == len(distribution.learning_set)
    else:
        assert distribution.method == distribution.CROSS_VALIDATION
        assert distribution.size == size or distribution.N_FOLDS

    assert len(distribution.algos) == distribution.size
    for algo in distribution.algos:
        assert isinstance(algo, LinearRegressor)


def test_learn(distribution):
    """Check the results of the learning stage for the original model.

    Original model: f(x) = 2/3
    Sub-models: f0(x) = -1 + 2x, f1(x) = 1, f2(x) = 1 - 2x
    """
    assert pytest.approx(distribution.algo.intercept[0], 0.1) == 2.0 / 3
    assert pytest.approx(distribution.algo.coefficients[0], 0.1) == 0.0


@pytest.mark.parametrize(
    ("model", "intercept", "coefficient"), [(0, -1, 2.0), (1, 1.0, 0.0), (2, 1.0, -2.0)]
)
def test_learn_submodels(distribution, model, intercept, coefficient):
    """Check the results of the learning stage for the sub-models.

    Sub-models: f0(x) = -1 + 2x, f1(x) = 1, f2(x) = 1 - 2x
    """
    assert intercept == pytest.approx(distribution.algos[model].intercept[0], 0.1)
    assert coefficient == pytest.approx(distribution.algos[model].coefficients[0], 0.1)


@pytest.mark.parametrize("point", [0.0, 0.25, 0.5, 0.75, 1.0])
@pytest.mark.parametrize("model", [0, 1, 2])
def test_weight_functions(distribution, point, model):
    """Check the weight functions."""
    func = distribution.weights[model]
    assert __weight_func(point, [model]) == func(array([point]))
    assert __weight_func(point, [model]) == func({"x": array([point])})


@pytest.mark.parametrize("model", [0, 1, 2])
def test_weight_function_with_several_points(distribution, model):
    """Check the weight functions with several points at once."""
    points = array([0.0, 0.25, 0.5, 0.75, 1.0])
    expected = array([__weight_func(point, [model]) for point in points])
    returned = distribution.weights[model](points.reshape((-1, 1)))
    assert returned.shape == (5,)
    assert sum(abs(expected - returned)) == 0


@pytest.mark.parametrize("point", [0.0, 0.25, 0.5, 0.75, 1.0])
def test_evaluate_weights(distribution, point):
    """Check the evaluate function."""
    assert distribution.evaluate_weights(array([point])).shape == (3,)


def test_evaluate_weights_with_several_points(distribution):
    """Check the evaluate function with several points at once."""
    weights = distribution.evaluate_weights(array([[0.0, 0.25, 0.5, 0.75, 1.0]]).T)
    assert weights.shape == (3, 5)
    for value in weights.sum(0):
        assert pytest.approx(value, 0.1) == 1


@pytest.mark.parametrize(
    ("point", "outputs"),
    [
        (array([0.0]), [-1, 1, 1]),
        ({"x": array([0.0])}, [-1, 1, 1]),
        (array([[0.0], [1.0]]), [[-1, 1, 1], [1, 1, -1]]),
        ({"x": array([[0.0], [1.0]])}, [[-1, 1, 1], [1, 1, -1]]),
    ],
)
def test_predict_members(distribution, point, outputs):
    """Check the predictions of the members."""
    predictions = distribution.predict_members(point)
    if isinstance(point, dict):
        predictions = predictions["y"]
        point = point["x"]

    n_points = point.shape[0]
    assert predictions.shape == (3, 1) if n_points == 1 else (3, n_points, 1)
    if n_points == 1:
        for model in [0, 1, 2]:
            assert predictions[model, 0] == pytest.approx(outputs[model], 0.1)
    else:
        for model in [0, 1, 2]:
            for point in range(n_points):
                expected = pytest.approx(outputs[point][model], 0.1)
                assert predictions[model, point, 0] == expected


@pytest.mark.parametrize(
    "point",
    [
        array([0.0]),
        array([0.5]),
        array([1.0]),
        array([[0.0], [0.5]]),
        {"x": array([0.0])},
        {"x": array([[0.0], [0.5]])},
    ],
)
def test_compute_mean(distribution, point):
    """Check the computation of the mean."""
    result = distribution.compute_mean(point)
    if isinstance(point, dict):
        point = point["x"]
        result = result["y"]

    assert result.shape == point.shape
    assert result[0] == pytest.approx(0 if point[0] == 0.5 else 1, 0.1)


@pytest.mark.parametrize(
    "point",
    [
        array([0.0]),
        array([0.5]),
        array([1.0]),
        array([[0.0], [0.5]]),
        {"x": array([0.0])},
        {"x": array([[0.0], [0.5]])},
    ],
)
def test_compute_variance(distribution, point):
    """Check the computation of the variance."""
    result = distribution.compute_variance(point)
    if isinstance(point, dict):
        point = point["x"]
        result = result["y"]

    assert result.shape == point.shape
    assert result[0] == pytest.approx(0, 0.1)


@pytest.mark.parametrize(
    ("point", "outputs"),
    [
        (array([0.0]), [array([-1, 1, 1])]),
        (array([1.0]), [array([-1, 1, 1])]),
        (
            array([[0.0], [0.5], [1.0]]),
            [array([-1, 1, 1]), array([0, 0, 1]), array([-1, 1, 1])],
        ),
        ({"x": array([0.0])}, [array([-1, 1, 1])]),
        (
            {"x": array([[0.0], [0.5], [1.0]])},
            [array([-1, 1, 1]), array([0, 0, 1]), array([-1, 1, 1])],
        ),
    ],
)
def test_compute_confidence_interval(distribution, point, outputs):
    """Check the computation of a confidence interval."""
    lower, upper = distribution.compute_confidence_interval(point)
    if isinstance(point, dict):
        point = point["x"]
        lower = lower["y"]
        upper = upper["y"]

    assert lower.shape == upper.shape == point.shape

    for index, output in enumerate(outputs):
        assert pytest.approx(lower[index], 0.1) == quantile(output, 0.025)
        assert pytest.approx(upper[index], 0.1) == quantile(output, 0.975)


@pytest.mark.parametrize(
    "input_data",
    [
        array([[1 / 3.97]]),
        array([[1 / 3.97], [1 / 1.97]]),
    ],
)
def test_compute_samples(distribution, input_data):
    """Check that the samples from the regressor are of appropriate shape."""
    nb_samples = distribution.size
    samples = distribution.compute_samples(input_data=input_data, n_samples=nb_samples)
    assert samples.shape == (nb_samples, input_data.shape[0], 1)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The settings for the SMT's regression models."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Final

from gemseo.mlearning.regression.algos.base_regressor_settings import (
    BaseRegressorSettings,
)
from pydantic import Field
from pydantic import field_validator
from smt.surrogate_models.surrogate_model import SurrogateModel
from strenum import StrEnum

if TYPE_CHECKING:
    from collections.abc import Iterator

    from gemseo import StrKeyMapping


def _get_subclasses(cls: type[SurrogateModel]) -> Iterator[type[SurrogateModel]]:
    """Return the subclasses of a class of interest recursively.

    Args:
        cls: The class of interest.

    Yields:
        The subclasses of the class of interest.
    """
    for subclass in cls.__subclasses__():
        yield from _get_subclasses(subclass)
        yield subclass


_NAMES_TO_CLASSES: Final[dict[str, type[SurrogateModel]]] = {
    cls.__name__: cls for cls in _get_subclasses(SurrogateModel)
}
"""The model class names bound to the model classes."""

SMTSurrogateModel = StrEnum("SurrogateModel", list(_NAMES_TO_CLASSES.keys()))
"""The class name of an SMT surrogate model."""


class SMT_Regressor_Settings(BaseRegressorSettings):  # noqa: N801
    """The settings for the SMT's regression models."""

    model_class_name: SMTSurrogateModel = Field(
        default=...,
        description=(
            "The class name of a surrogate model available in SMT, "
            "i.e. a subclass of `smt.surrogate_models.surrogate_model.SurrogateModel`."
        ),
    )

    @field_validator("parameters")
    @classmethod
    def __set_print_global(cls, parameters: StrKeyMapping) -> StrKeyMapping:
        """Set `print_global` if missing."""
        new_parameters = {"print_global": False}
        new_parameters.update(parameters)
        return new_parameters


# TODO: API: remove this alias.
SMTRegressorSettings = SMT_Regressor_Settings



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Distance to the learning set."""

from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import atleast_2d
from numpy import nonzero
from scipy.spatial.distance import cdist

from gemseo_mlearning.active_learning.acquisition_criteria.exploration.base_exploration import (  # noqa: E501
    BaseExploration,
)

if TYPE_CHECKING:
    from gemseo.typing import NumberArray


class Distance(BaseExploration):
    """Distance to the learning set.

    This acquisition criterion computes the minimum distance between a new point and the
    point of the learning dataset, scaled by the minimum distance between two distinct
    learning points.
    """

    def _compute(self, input_value: NumberArray) -> NumberArray | float:  # noqa: D102
        train = self._regressor_distribution.learning_set.input_dataset.to_numpy()
        return (
            cdist(atleast_2d(input_value), train).min(axis=-1)
            / (dist_train := cdist(train, train))[nonzero(dist_train)].min()
            * 2.0
        )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""History view."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from gemseo.datasets.io_dataset import IODataset
from gemseo.post.dataset.lines import Lines
from numpy import array
from numpy import newaxis

if TYPE_CHECKING:
    from pathlib import Path

    from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo


class QOIHistoryView:
    """View of the history of the quantity of interest (QOI)."""

    __algo: ActiveLearningAlgo
    """The active learning algorithm."""

    def __init__(self, active_learning_algo: ActiveLearningAlgo) -> None:
        """
        Args:
            active_learning_algo: The active learning algorithm.

        Raises:
            ValueError: When there is no quantity of interest
                associated with the acquisition criterion.
        """  # noqa: D205, D212
        if active_learning_algo.acquisition_criterion.qoi is None:
            msg = (
                "There is no quantity of interest "
                "associated with the acquisition criterion "
                f"{active_learning_algo.acquisition_criterion.__class__.__name__}."
            )
            raise ValueError(msg)

        self.__algo = active_learning_algo

    def draw(
        self,
        show: bool = True,
        file_path: str | Path = "",
        label: str = "",
        add_markers: bool = True,
        **options: Any,
    ) -> Lines:
        """Draw the QOI history.

        Args:
            show: Whether to display the view.
            file_path: The file path to save the view.
                If empty, do not save it.
            label: The label for the QOI.
                If empty, use the name of the acquisition criterion family.
            add_markers: Whether to add markers.
            **options: The options to create the
                [Lines][gemseo.post.dataset.lines.Lines] object.

        Returns:
            A view of the QOI history.
        """
        if not label:
            label = self.__algo.acquisition_criterion_family_name

        x_label = "Number of evaluations"
        n_evaluations_history, qoi_history = self.__algo.qoi_history
        qoi_history = [qoi[0] for qoi in qoi_history]
        dataset = IODataset()
        dataset.add_variable(x_label, array(n_evaluations_history)[:, newaxis])
        dataset.add_variable(label, array(qoi_history)[:, newaxis])
        lines = Lines(
            dataset,
            variables=[label],
            abscissa_variable=x_label,
            add_markers=add_markers,
            **options,
        )
        lines.marker = "."
        lines.execute(show=show, save=file_path != "", file_path=file_path)
        return lines



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Family of acquisition criteria to estimate a level set."""

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion_family import (  # noqa: E501
    BaseAcquisitionCriterionFamily,
)
from gemseo_mlearning.active_learning.acquisition_criteria.level_set.factory import (
    LevelSetFactory,
)


class LevelSet(BaseAcquisitionCriterionFamily):
    """The family of acquisition criteria to estimate a level set."""

    ACQUISITION_CRITERION_FACTORY = LevelSetFactory



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

from gemseo_mlearning.problems.branin.branin_function import BraninFunction
from gemseo_mlearning.problems.branin.functions import compute_gradient
from gemseo_mlearning.problems.branin.functions import compute_output


def test_branin_function() -> None:
    """Check that the BraninFunction has the right name, function and jacobian."""
    function = BraninFunction()
    assert function.name == "Branin"
    assert function._func == compute_output
    assert function._jac == compute_gradient



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Base class for acquisition criteria to estimate a minimum."""

from __future__ import annotations

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion import (  # noqa: E501
    BaseAcquisitionCriterion,
)


class BaseMinimum(BaseAcquisitionCriterion):
    """The base class for acquisition criteria to estimate a minimum."""

    def update(self) -> None:  # noqa: D102
        super().update()
        self._qoi = min(
            self._regressor_distribution.learning_set.output_dataset.to_numpy()
        )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Base class for the acquisition criteria to explore the input space."""

from __future__ import annotations

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion import (  # noqa: E501
    BaseAcquisitionCriterion,
)


class BaseExploration(BaseAcquisitionCriterion):
    """The base class the for acquisition criteria to explore the input space.

    Optimize it to make the error of the surrogate model tends towards zero when the
    number of acquired points tends to infinity.
    """



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import re

import pytest
from gemseo.algos.parameter_space import ParameterSpace
from numpy import array
from numpy.testing import assert_almost_equal

from gemseo_mlearning.active_learning.acquisition_criteria.quantile.ef import EF
from gemseo_mlearning.active_learning.acquisition_criteria.quantile.ei import EI
from gemseo_mlearning.active_learning.acquisition_criteria.quantile.u import U


@pytest.fixture(scope="module")
def uncertain_space() -> ParameterSpace:
    """An uncertain space."""
    parameter_space = ParameterSpace()
    parameter_space.add_random_variable("a", "OTNormalDistribution")
    parameter_space.add_random_variable("x", "OTNormalDistribution")
    return parameter_space


@pytest.mark.parametrize(
    ("cls", "input_value", "expected"),
    [
        (U, array([0.25]), array([8.5598195e-14])),
        (EF, array([0.25]), array([0.5746878])),
        (EI, array([0.25]), array([0.6843251])),
        (U, array([[0.25]] * 2), array([[8.5598195e-14]] * 2)),
        (EF, array([[0.25]] * 2), array([[0.5746878]] * 2)),
        (EI, array([[0.25]] * 2), array([[0.6843251]] * 2)),
    ],
)
def test_quantile_kriging(
    kriging_distribution, uncertain_space, cls, input_value, expected
):
    """Check the criteria deriving from BaseQuantile with a Kriging distribution."""
    criterion = cls(kriging_distribution, 0.8, uncertain_space)
    assert_almost_equal(criterion.func(input_value), expected)


@pytest.mark.parametrize(
    ("cls", "input_value", "expected"),
    [
        (U, array([0.25]), array([0.11785113])),
        (EF, array([0.25]), array([0.25])),
        (EI, array([0.25]), array([0.64])),
        (U, array([[0.25]] * 2), array([[0.11785113]] * 2)),
        (EF, array([[0.25]] * 2), array([[0.25]] * 2)),
        (EI, array([[0.25]] * 2), array([[0.64]] * 2)),
    ],
)
def test_quantile_regressor(
    algo_distribution, cls, input_value, expected, uncertain_space
):
    """Check the criteria deriving from BaseQuantile with non GP regressor."""
    criterion = cls(
        regressor_distribution=algo_distribution,
        level=0.8,
        uncertain_space=uncertain_space,
    )
    assert_almost_equal(criterion.func(input_value), expected, decimal=2)


@pytest.mark.parametrize(
    ("cls", "input_value", "expected"),
    [
        (U, array([[0.25]] * 2), array([1.0013774e-08])),
        (EF, array([[0.25]] * 2), 0.9428065),
        (EI, array([[0.25]] * 2), array([0.8888889])),
    ],
)
def test_quantile_parallel(
    kriging_distribution, cls, input_value, expected, uncertain_space
):
    """Check the parallelized criteria deriving from BaseQuantile."""
    criterion = cls(
        regressor_distribution=kriging_distribution,
        level=0.8,
        uncertain_space=uncertain_space,
        batch_size=2,
    )
    assert_almost_equal(criterion.func(input_value), expected)


def test_quantile_error(algo_distribution):
    """Check the exception raised by a BaseQuantile criterion."""
    uncertain_space = ParameterSpace()
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The probability distributions of the input variables x are missing."
        ),
    ):
        EF(algo_distribution, 0.8, uncertain_space)


@pytest.mark.parametrize(
    "cls",
    [
        U,
        EI,
        EF,
    ],
)
def test_bad_parallel_regressor(algo_distribution, cls, uncertain_space):
    """Check that parallelized criteria with non GP regressor lead to failure."""
    criterion = cls(
        regressor_distribution=algo_distribution,
        level=0.8,
        uncertain_space=uncertain_space,
        batch_size=2,
    )

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "Parallelization with batch_size > 1 is not yet implemented "
            "for regressors that are not based on a random process."
        ),
    ):
        criterion.func(array([0.123]))



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import annotations

import pytest
from gemseo.algos.doe.factory import DOELibraryFactory
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from gemseo.problems.optimization.rastrigin import Rastrigin


@pytest.fixture
def regression_algorithm() -> OTGaussianProcessRegressor:
    """A regression algorithm for the Rastrigin problem."""
    problem = Rastrigin()
    DOELibraryFactory().execute(problem, algo_name="OT_SOBOL", n_samples=5)
    dataset = problem.to_dataset(opt_naming=False)
    dataset = dataset.map(lambda x: x.real)
    algo = OTGaussianProcessRegressor(
        dataset, transformer=OTGaussianProcessRegressor.DEFAULT_TRANSFORMER
    )
    algo.learn()
    return algo
